import re
import socket

from sauna.plugins import (Plugin, bytes_to_human, human_to_bytes,
                           PluginRegister)

my_plugin = PluginRegister('Memcached')


@my_plugin.plugin()
class Memcached(Plugin):

    def __init__(self, config):
        super().__init__(config)
        self.config = {
            'host': config.get('host', 'localhost'),
            'port': config.get('port', 11211),
            'timeout': config.get('timeout', 5)
        }
        self._stats = None

    @my_plugin.check()
    def accepting_connections(self, check_config):
        try:
            accept_connections = self.stats['accepting_conns'] == 1
        except OSError as e:
            return (Plugin.STATUS_CRIT,
                    'Memcached is not accepting connections: {}'.format(e))
        if accept_connections:
            return Plugin.STATUS_OK, 'Memcached is accepting connections'
        else:
            return Plugin.STATUS_CRIT, 'Memcached is not accepting connections'

    @my_plugin.check()
    def bytes(self, check_config):
        status = self._value_to_status_less(self.stats['bytes'], check_config,
                                            human_to_bytes)
        output = 'Memcached memory: {}'.format(
            bytes_to_human(self.stats['bytes'])
        )
        return status, output

    @my_plugin.check()
    def used_percent(self, check_config):
        used_percent = int(
            self.stats['bytes'] * 100 / self.stats['limit_maxbytes']
        )
        status = self._value_to_status_less(used_percent, check_config,
                                            self._strip_percent_sign)
        output = 'Memcached memory used: {}% of {}'.format(
            used_percent, bytes_to_human(self.stats['limit_maxbytes'])
        )
        return status, output

    @my_plugin.check()
    def current_items(self, check_config):
        status = self._value_to_status_less(self.stats['curr_items'],
                                            check_config)
        output = 'Memcached holds {} items'.format(self.stats['curr_items'])
        return status, output

    @property
    def stats(self):
        if not self._stats:
            self._stats = self._raw_stats_to_dict(
                self._fetch_memcached_stats()
            )
        return self._stats

    @classmethod
    def _raw_stats_to_dict(cls, stats_data):
        """Convert raw memcached output to a dict of stats."""
        stats_string = stats_data.decode('ascii')
        stats_string = stats_string.replace('\r\n', '\n')
        matches = re.findall(r'^STAT (\w+) (\d+)$', stats_string,
                             flags=re.MULTILINE)
        return {match[0]: int(match[1]) for match in matches}

    def _fetch_memcached_stats(self):
        """Connect to Memcached and retrieve stats."""
        data = bytes()
        with socket.create_connection((self.config['host'],
                                       self.config['port']),
                                      timeout=self.config['timeout']) as s:
            s.sendall(b'stats\r\n')
            while True:
                buffer = bytearray(4096)
                bytes_received = s.recv_into(buffer)
                if bytes_received == 0:
                    # Remote host closed connection
                    break
                data += buffer
                if b'\r\nEND\r\n' in data:
                    # End of the stats command
                    break
        return data

    @staticmethod
    def config_sample():
        return '''
        # Memcached
        - type: Memcached
          checks:
            - type: bytes
              warn: 128M
              crit: 256M
            - type: used_percent
              warn: 80%
              crit: 90%
            - type: current_items
              warn: 10000
              crit: 20000
            - type: accepting_connections
          config:
            host: localhost
            port: 11211
        '''



import os
import glob
from collections import namedtuple
from functools import reduce

from sauna.plugins import Plugin, PluginRegister


Sensor = namedtuple('Sensor', ['device_name', 'label', 'value'])

my_plugin = PluginRegister('Hwmon')


@my_plugin.plugin()
class Hwmon(Plugin):
    """Linux hardware monitoring plugin.

    This plugin crawls Linux's /sys/class/hwmon to find usable sensors. Be
    warned that this method is quite fragile since exotic hardware may present
    values that need offsets or conversions.

    A more solid approach could be to use lm-sensors, but:
    - it requires to install and configure lm-sensors
    - there is no proper python bindings to the library
    - parsing the output of 'sensors' is not fun nor efficient
    """

    @my_plugin.check()
    def temperature(self, check_config):
        dummy_sensor = Sensor(device_name='Dummy', label='Dummy', value=-1000)
        sensors = self._get_temperatures()
        if check_config.get('sensors'):
            sensors = [
                sensor for sensor in sensors
                if sensor.device_name in check_config.get('sensors', [])
            ]
        sensor = reduce(lambda x, y: x if x.value > y.value else y,
                        sensors,
                        dummy_sensor)
        if sensor is dummy_sensor:
            return self.STATUS_UNKNOWN, 'No sensor found'
        status = self._value_to_status_less(sensor.value, check_config)
        if status > self.STATUS_OK:
            return (
                status,
                'Sensor {}/{} {}°C'.format(sensor.device_name,
                                           sensor.label,
                                           sensor.value)
            )
        return self.STATUS_OK, 'Temperature okay ({}°C)'.format(sensor.value)

    @classmethod
    def _get_temperatures(cls):
        temperatures = list()
        for device in cls._get_devices():
            temperatures.extend(cls._process_device(device))
        return temperatures

    @classmethod
    def _get_devices(cls):
        base_path = '/sys/class/hwmon'
        devices = set()
        for file in os.listdir(base_path):
            file_path = os.path.join(base_path, file)
            if os.path.isfile(os.path.join(file_path, 'device', 'name')):
                devices.add(os.path.join(file_path, 'device'))
            else:
                devices.add(file_path)
        return devices

    @classmethod
    def _process_device(cls, device):
        sensors = list()
        with open(os.path.join(device, 'name')) as f:
            device_name = f.read().strip()
        pattern = 'temp*_input'
        for temp_file in glob.glob(os.path.join(device, pattern)):
            with open(temp_file) as f:
                temperature = int(int(f.read().strip())/1000)
            try:
                with open(temp_file.replace('input', 'label')) as f:
                    label = f.read().strip()
            except (OSError, IOError):
                label = None

            sensors.append(Sensor(device_name, label, temperature))

        return sensors

    @staticmethod
    def config_sample():
        return '''
        # Linux hardware monitoring
        - type: Hwmon
          checks:
            # Raise an alert if any sensor gets beyond a threshold
            - type: temperature
              warn: 65
              crit: 85
              # Optionally, only check some sensors
              sensors: ['acpitz', 'coretemp']
        '''
